package com.brilliance.mda.support.mybatis;

import com.brilliance.mda.runtime.mda.IModule;
import com.brilliance.mda.runtime.mda.impl.Argument;
import com.brilliance.mda.support.mybatis.entity.Column;
import com.brilliance.mda.support.mybatis.entity.Table;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.util.StringUtils;

import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import static com.brilliance.mda.runtime.mda.OpType.*;

/**
 * @author hulei
 * bd代码中sql语句与mybatis sql之间的适配转换器
 */
public class MybatisArgumentAdapter {

    private static Map<String, String> annotCache = new ConcurrentHashMap<>();

    public static final String HOLDER_SQL = "sql";
    public static final String HOLDER_CONDS = "conditions";

    public static enum AdaType {

        SELECT("SELECT * FROM %s WHERE %s"),
        DELETE("DELETE   FROM %s WHERE %s");

        private String value;

        private AdaType(String val) {
            this.value = val;
        }

        public String getValue() {
            return value;
        }
    }

    private Logger log = LoggerFactory.getLogger(this.getClass());

    private Class<? extends IModule> clazz;
    private List<Class<? extends IModule>> clazzList;
    private Argument[] args;
    private String sqlTemplate;
    private Map<String, Object> sqlParams;
    private int idx = 0;
    private String whereSql;
    private String pattern;
    private Object[] params;
    private String tableName;


    public MybatisArgumentAdapter(AdaType type, Class<? extends IModule> clazz, Argument... args) {
        this.clazz = clazz;
        this.pattern = type.getValue();
        this.args = args;
        this.tableName = convertTableName();
        this.init();
    }

    public MybatisArgumentAdapter(AdaType type, Class<? extends IModule> clazz, String whereSql) {
        this(type, clazz, whereSql, null);
    }

    public MybatisArgumentAdapter(AdaType type, Class<? extends IModule> clazz, String whereSql, Object[] params) {
        this.clazz = clazz;
        this.pattern = type.getValue();
        this.whereSql = whereSql;
        this.params = params;
        this.tableName = convertTableName();
        this.init();
    }

    public MybatisArgumentAdapter(AdaType type, List<Class<? extends IModule>> clazzList, String whereSql, Object[] params) {
        this.clazzList = clazzList;
        this.pattern = type.getValue();
        this.whereSql = whereSql;
        this.params = params;
        this.tableName = convertTableNames();
        this.init();
    }


    public void init() {

        sqlParams = new HashMap<>();

        String conditions = "";
        if (whereSql != null && !whereSql.trim().equals("")) {
            if (whereSql.toUpperCase().indexOf("WHERE") > -1) {
                pattern = pattern.replace("WHERE", "");
            }
            if (params != null && params.length > 0) {
                conditions = convertPlaceHolder(whereSql, params, sqlParams);
            } else {
                conditions = whereSql;
            }
        } else {
            if (this.args != null && this.args.length > 0) {
                conditions = parserArgument(this.args);
            } else {
                conditions = " 1=1 ";
            }
        }

        this.sqlTemplate = String.format(pattern, tableName, conditions);
        this.sqlParams.put(HOLDER_SQL, this.sqlTemplate);
        this.sqlParams.put(HOLDER_CONDS, conditions);
    }


    public String getSqlTemplate() {
        return sqlTemplate;
    }

    public Map<String, Object> getSqlParams() {
        return sqlParams;
    }

    /**
     * 获取直接WHERE条件进行参数替换
     */
    protected final String parserArgument(String whereClause, Object[] args) {
        StringBuffer sb = new StringBuffer(" ");
        Matcher m = Pattern.compile("\\?").matcher(whereClause);
        int offset = 0, i = 0;
        while (m.find()) {
            sb.append(whereClause.substring(offset, m.start()));
            sb.append("#{args[" + i + "]}");
            i++;
            offset = m.end();
        }
        sb.append(whereClause.substring(offset));
        if (log.isDebugEnabled())
            log.debug(sb.toString());
        return sb.toString();
    }

    /**
     * 获取生成对应的SQL条件（这里完成最终的组装）
     */
    protected final String parserArgument(Argument... arguments) {
        List<String> orders = new ArrayList();
        List<String> criterions = new ArrayList();
        for (int i = 0; i < arguments.length; i++)
            process(arguments[i], i, orders, criterions);
        StringBuffer sb = new StringBuffer();
        String conn = "";
        for (String criterion : criterions) {
            sb.append(conn).append(criterion);
            conn = " AND ";
        }
        conn = " ORDER BY ";
        for (String criterion : orders) {
            sb.append(conn).append(criterion);
            conn = ",";
        }
        if (log.isDebugEnabled())
            log.debug(sb.toString());
        return sb.toString();
    }

    /**
     * 按Order和Criterion来处理Argument
     */
    private final void process(Argument argument, int index, List orders, List criterions) {
        if (argument.opType == ASC)
            orders.add(argument.fieldName);
        else if (argument.opType == DESC)
            orders.add(argument.fieldName + " DESC");
        else {
            String criterion = processCriterion(argument, index);
            if (criterion != null)
                criterions.add(criterion);
        }
    }

    private final String processCriterion(Argument argument, int idx) {
        if (argument == null || argument.opType == null)
            return null;
        StringBuffer sb;
        String conn;
        String holderName;
        switch (argument.opType) {
            case ISNULL:
                return convertColumn(argument.fieldName) + " IS NULL";
            case ISNOTNULL:
                return convertColumn(argument.fieldName) + " IS NOT NULL";
            case NE:
                holderName = argument.fieldName + "_" + this.idx++;
                sqlParams.put(holderName, argument.value);
                return convertColumn(argument.fieldName) + "!=" + placeholder(argument.value, holderName);
            case LE:
                holderName = argument.fieldName + "_" + this.idx++;
                sqlParams.put(holderName, argument.value);
                return convertColumn(argument.fieldName) + "<=" + placeholder(argument.value, holderName);
            case GE:
                holderName = argument.fieldName + "_" + this.idx++;
                sqlParams.put(holderName, argument.value);
                return convertColumn(argument.fieldName) + ">=" + placeholder(argument.value, holderName);
            case LT:
                holderName = argument.fieldName + "_" + this.idx++;
                sqlParams.put(holderName, argument.value);
                return convertColumn(argument.fieldName) + "<" + placeholder(argument.value, holderName);
            case GT:
                holderName = argument.fieldName + "_" + this.idx++;
                sqlParams.put(holderName, argument.value);
                return convertColumn(argument.fieldName) + ">" + placeholder(argument.value, holderName);
            case NOTLIKE:
                holderName = argument.fieldName + "_" + this.idx++;
                sqlParams.put(holderName, argument.value);
                return convertColumn(argument.fieldName) + " NOT LIKE " + placeholder(argument.value, holderName);
            case LIKE:
                holderName = argument.fieldName + "_" + this.idx++;
                sqlParams.put(holderName, argument.value);
                return convertColumn(argument.fieldName) + " LIKE " + placeholder(argument.value, holderName);
            case BETWEEN:
                holderName = argument.fieldName + "_" + this.idx++;
                Object[] ary = (Object[]) argument.value;
                sqlParams.put(holderName + "_0", ary[0]);
                sqlParams.put(holderName + "_1", ary[1]);
                return convertColumn(argument.fieldName) + " BETWEEN " + placeholder(ary[0], holderName + "_0") + " AND " + placeholder(ary[1], holderName + "_1");
            case IN:
                holderName = argument.fieldName + "_" + this.idx++;
                sb = new StringBuffer(convertColumn(argument.fieldName)).append(" IN (");
                conn = "";
                Object[] values = (Object[]) argument.value;
                for (int i = 0; i < values.length; i++) {
                    sqlParams.put(holderName + "_" + i, values[i]);
                    sb.append(conn).append(placeholder(values[i], holderName + "_" + i));
                    conn = ",";
                }
                return sb.append(")").toString();
            case AND:
            case OR:
                sb = new StringBuffer("(");
                conn = "";
                Argument[] arguments = (Argument[]) argument.value;
                for (int i = 0; i < arguments.length; i++) {
                    sb.append(conn).append(processCriterion(arguments[i], i));
                    conn = argument.opType == AND ? " AND " : " OR ";
                }
                return sb.append(")").toString();
            case NOT:
                return "!(" + processCriterion((Argument) argument.value, idx) + ")";
            default:
                holderName = argument.fieldName + "_" + this.idx++;
                sqlParams.put(holderName, argument.value);
                return convertColumn(argument.fieldName) + "=" + placeholder(argument.value, holderName);
        }
    }

    private final String placeholder(Object value, String name) {
        return String.format("#{%s}", name);
    }


//    /**
//     * 获取指定属性名中@Column注解的值
//     * @param clazz
//     * @param name
//     * @return
//     */
//    public static String getFieldAnnotation(Class<? extends IModule> clazz,String name){
//
//    }

    /**
     * 将标准SQL占位符替换为mybatis格式的占位符，并将参数数组转换为HashMap
     *
     * @param sql
     * @param objects
     * @param map
     * @return
     */
    public static String convertPlaceHolder(String sql, Object[] objects, Map<String, Object> map) {

        Pattern p = Pattern.compile("[?]");
        Matcher m = p.matcher(sql);

        StringBuffer buf = new StringBuffer();
        for (int i = 0; m.find() && i < objects.length; i++) {
            String holder = "var_" + i;
            m.appendReplacement(buf, String.format("#{%s}", holder));
            map.put(holder, objects[i]);
        }
        m.appendTail(buf);
        return buf.toString();
    }

    public String convertTableName() {
        if (annotCache.containsKey(clazz.getName())) {
            return annotCache.get(clazz.getName());
        }
        String value = clazz.getSimpleName();
        if (clazz.isAnnotationPresent(Table.class)) {
            Table t = clazz.getAnnotation(Table.class);
            if (t.value() != null) {
                value = t.value();
            }
        }
        annotCache.put(clazz.getName(), value);
        return value;
    }


    private String convertTableNames() {
        StringBuffer sb = new StringBuffer("");
        for (Class c : clazzList) {
            sb.append(c.getName());
        }
        if (annotCache.containsKey(sb.toString())) {
            return annotCache.get(sb.toString());
        }

        List<String> tb = new ArrayList<>();
        for (int i = 0; i < clazzList.size(); i++) {
            int idx = i + 1;
            Class<? extends IModule> c = clazzList.get(i);
            String value = c.getSimpleName();
            if (c.isAnnotationPresent(Table.class)) {
                Table t = c.getAnnotation(Table.class);
                if (t.value() != null) {
                    value = t.value();
                }
            }
            tb.add(value + " m" + idx);
        }
        String value = String.join(",", tb);
        annotCache.put(sb.toString(), value);
        return value;
    }


    public String convertColumn(String fieldName) {
        String key = String.format("%s.%s", clazz.getName(), fieldName);
        if (annotCache.containsKey(key)) {
            return annotCache.get(key);
        }

        try {
            Field field = clazz.getDeclaredField(fieldName);
            String value = fieldName;
            if (field.isAnnotationPresent(Column.class)) {
                Column col = field.getAnnotation(Column.class);
                if (!StringUtils.isEmpty(col.value())) {
                    value = col.value();
                }
            }
            annotCache.put(key, value);
        } catch (NoSuchFieldException e) {
            annotCache.put(key, fieldName);
            return fieldName;
        }
        return annotCache.get(key);
    }

}
